import os
from typing import Callable, Union

import cloudpickle
import numpy as np
import pandas as pd

from clearbox_wrapper.exceptions import ClearboxWrapperException

DataPreparationInput = Union[pd.DataFrame, pd.Series, np.ndarray]
DataPreparationOutput = Union[pd.DataFrame, pd.Series, np.ndarray]


class DataPreparation(object):
    def __init__(self, data_preparation_function: Callable):
        """Create a DataPreparation instance.

        Parameters
        ----------
        preprocessing_function : Callable
            A function to use as a preprocessor. You can use your own custom code for
            preprocessing, but it must be wrapped in a single function.

            NOTE: If the preprocessing includes any kind of fitting on the training dataset
            (e.g. Scikit Learn transformers), it must be performed outside the final
            preprocessing function to save. Fit the transformer(s) outside the function and
            put only the transform method inside it. Furthermore, if the entire preprocessing
            is performed with a single Scikit-Learn transformer, you can directly pass it
            (fitted) to this method.

        Raises
        ------
        TypeError
            If preprocessing_function is not a function (Callable type)
        """
        self.data_preparation = data_preparation_function

    def __repr__(self) -> str:
        return "Data Preparation: \n" "  {}\n".format(repr(self.data_preparation))

    @property
    def data_preparation(self) -> Callable:
        """Get the data preparation function.

        Returns
        -------
        Callable
            The data preparation function.
        """
        return self._data_preparation

    @data_preparation.setter
    def data_preparation(self, data_preparation_function: Callable) -> None:
        """Set the data preparation function.

        Parameters
        ----------
        value : Callable
            The data preparation function.
        """
        self._data_preparation = data_preparation_function

    def prepare_data(self, data: DataPreparationInput) -> DataPreparationOutput:
        """Prepare input data using the data preparation function.

        Parameters
        ----------
        data : DataPreparationInput
            Input data to prepare.

        Returns
        -------
        DataPreparationOutput
            Prepared data.
        """
        prepared_data = (
            self.data_preparation.transform(data)
            if hasattr(self.data_preparation, "transform")
            else self.data_preparation(data)
        )
        return prepared_data

    def save(self, path: str) -> None:
        if os.path.exists(path):
            raise ClearboxWrapperException(
                "Data preparation path '{}' already exists".format(path)
            )
        with open(path, "wb") as data_preparation_serialized_file:
            cloudpickle.dump(self, data_preparation_serialized_file)


def create_and_save_data_preparation(
    data_preparation_function: Callable, path: str
) -> None:
    """Create, serialize and save a DataPreparation instance.

    Parameters
    ----------
    data_preparation_function : Callable
        A function to use as data preparation. You can use your own custom code for
        data preparation, but it must be wrapped in a single function.

        NOTE: If the data preparation includes any kind of fitting on the training dataset
        (e.g. Scikit Learn transformers), it must be performed outside the final data
        preparation function to save. Fit the transformer(s) outside the function and put
        only the transform method inside it. Furthermore, if the entire data preparation
        is performed with a single Scikit-Learn transformer, you can directly pass it
        (fitted) to this method.
    path : str
        Local path to save the data preparation to.

    Raises
        ------
        TypeError
            If data_preparation_function is not a function (Callable type)
        ClearboxWrapperException
            If data preparation path already exists.
    """
    if not isinstance(data_preparation_function, Callable):
        raise TypeError(
            "data_preparation_function should be a Callable, got '{}'".format(
                type(data_preparation_function)
            )
        )
    if os.path.exists(path):
        raise ClearboxWrapperException(
            "Data preparation path '{}' already exists".format(path)
        )

    data_preparation = DataPreparation(data_preparation_function)
    with open(path, "wb") as data_preparation_serialized_file:
        cloudpickle.dump(data_preparation, data_preparation_serialized_file)


def load_serialized_data_preparation(
    serialized_data_preparation_path: str,
) -> DataPreparation:
    with open(serialized_data_preparation_path, "rb") as serialized_data_preparation:
        return cloudpickle.load(serialized_data_preparation)



"""
This module imports contents from CloudPickle in a way that is compatible with the
``pickle_module`` parameter of PyTorch's model persistence function: ``torch.save``
(see https://github.com/pytorch/pytorch/blob/692898fe379c9092f5e380797c32305145cd06e1/torch/
serialization.py#L192). It is included as a distinct module from :mod:`mlflow.pytorch` to avoid
polluting the namespace with wildcard imports.

Calling ``torch.save(..., pickle_module=mlflow.pytorch.pickle_module)`` will persist PyTorch
definitions using CloudPickle, leveraging improved pickling functionality such as the ability
to capture class definitions in the "__main__" scope.

TODO: Remove this module or make it an alias of CloudPickle when CloudPickle and PyTorch have
compatible pickling APIs.
"""

from cloudpickle import *  # noqa


from cloudpickle import CloudPickler as Pickler  # noqa


from pickle import Unpickler  # noqa



from .data_preparation import (
    create_and_save_data_preparation,
    DataPreparation,
    load_serialized_data_preparation,
)

__all__ = [
    create_and_save_data_preparation,
    load_serialized_data_preparation,
    DataPreparation,
]



import importlib
import os
import sys
from typing import Any, Dict, List, Union

from loguru import logger
import numpy as np
import pandas as pd
import yaml

from clearbox_wrapper.exceptions import ClearboxWrapperException
from clearbox_wrapper.model import MLMODEL_FILE_NAME, Model
from clearbox_wrapper.pyfunc import (
    add_pyfunc_flavor_to_model,
    CODE,
    DATA,
    FLAVOR_NAME,
    MAIN,
    PY_VERSION,
)
from clearbox_wrapper.pyfunc.model import (
    _save_model_with_class_artifacts_params,
    get_default_conda_env,
)
from clearbox_wrapper.schema import DataType, Schema
from clearbox_wrapper.signature import Signature
from clearbox_wrapper.utils import (
    _copy_file_or_tree,
    get_major_minor_py_version,
    PYTHON_VERSION,
)


PyFuncInput = Union[pd.DataFrame, np.ndarray, List[Any], Dict[str, Any]]
PyFuncOutput = Union[pd.DataFrame, pd.Series, np.ndarray, list]
new_model = Model()


def _enforce_type(name, values: pd.Series, t: DataType):
    """
    Enforce the input column type matches the declared in model input schema.
    The following type conversions are allowed:
    1. np.object -> string
    2. int -> long (upcast)
    3. float -> double (upcast)
    4. int -> double (safe conversion)
    Any other type mismatch will raise error.
    """
    if values.dtype == np.object and t not in (DataType.binary, DataType.string):
        values = values.infer_objects()

    if t == DataType.string and values.dtype == np.object:
        #  NB: strings are by default parsed and inferred as objects, but it is
        # recommended to use StringDtype extension type if available. See
        #
        # `https://pandas.pydata.org/pandas-docs/stable/user_guide/text.html`
        #
        # for more detail.
        try:
            return values.astype(t.to_pandas(), errors="raise")
        except ValueError:
            raise ClearboxWrapperException(
                "Failed to convert column {0} from type {1} to {2}.".format(
                    name, values.dtype, t
                )
            )

    # NB: Comparison of pandas and numpy data type fails when numpy data type is on the left
    # hand side of the comparison operator. It works, however, if pandas type is on the left
    # hand side. That is because pandas is aware of numpy.
    if t.to_pandas() == values.dtype or t.to_numpy() == values.dtype:
        # The types are already compatible => conversion is not necessary.
        return values

    if t == DataType.binary and values.dtype.kind == t.binary.to_numpy().kind:
        # NB: bytes in numpy have variable itemsize depending on the length of the longest
        # element in the array (column). Since MLflow binary type is length agnostic, we ignore
        # itemsize when matching binary columns.
        return values

    numpy_type = t.to_numpy()
    if values.dtype.kind == numpy_type.kind:
        is_upcast = values.dtype.itemsize <= numpy_type.itemsize
    elif values.dtype.kind == "u" and numpy_type.kind == "i":
        is_upcast = values.dtype.itemsize < numpy_type.itemsize
    elif values.dtype.kind in ("i", "u") and numpy_type == np.float64:
        # allow (u)int => double conversion
        is_upcast = values.dtype.itemsize <= 6
    else:
        is_upcast = False

    if is_upcast:
        return values.astype(numpy_type, errors="raise")
    else:
        # NB: conversion between incompatible types (e.g. floats -> ints or
        # double -> float) are not allowed. While supported by pandas and numpy,
        # these conversions alter the values significantly.
        def all_ints(xs):
            return all([pd.isnull(x) or int(x) == x for x in xs])

        hint = ""
        if (
            values.dtype == np.float64
            and numpy_type.kind in ("i", "u")
            and values.hasnans
            and all_ints(values)
        ):
            hint = (
                " Hint: the type mismatch is likely caused by missing values. "
                "Integer columns in python can not represent missing values and are therefore "
                "encoded as floats. The best way to avoid this problem is to infer the model "
                "schema based on a realistic data sample (training dataset) that includes "
                "missing values. Alternatively, you can declare integer columns as doubles "
                "(float64) whenever these columns may have missing values. See `Handling "
                "Integers With Missing Values <https://www.mlflow.org/docs/latest/models.html#"
                "handling-integers-with-missing-values>`_ for more details."
            )

        raise ClearboxWrapperException(
            "Incompatible input types for column {0}. "
            "Can not safely convert {1} to {2}.{3}".format(
                name, values.dtype, numpy_type, hint
            )
        )


def _enforce_schema(pdf: PyFuncInput, input_schema: Schema):
    """
    Enforce column names and types match the input schema.
    For column names, we check there are no missing columns and reorder the columns to match the
    ordering declared in schema if necessary. Any extra columns are ignored.
    For column types, we make sure the types match schema or can be safely converted to match
    the input schema.
    """
    if isinstance(pdf, (list, np.ndarray, dict)):
        try:
            pdf = pd.DataFrame(pdf)
        except Exception as e:
            message = (
                "This model contains a model signature, which suggests a DataFrame input."
                "There was an error casting the input data to a DataFrame: {0}".format(
                    str(e)
                )
            )
            raise ClearboxWrapperException(message)
    if not isinstance(pdf, pd.DataFrame):
        message = (
            "Expected input to be DataFrame or list. Found: %s" % type(pdf).__name__
        )
        raise ClearboxWrapperException(message)

    if input_schema.has_column_names():
        # make sure there are no missing columns
        col_names = input_schema.column_names()
        expected_names = set(col_names)
        actual_names = set(pdf.columns)
        missing_cols = expected_names - actual_names
        extra_cols = actual_names - expected_names
        # Preserve order from the original columns, since missing/extra columns are likely to
        # be in same order.
        missing_cols = [c for c in col_names if c in missing_cols]
        extra_cols = [c for c in pdf.columns if c in extra_cols]
        if missing_cols:
            message = (
                "Model input is missing columns {0}."
                " Note that there were extra columns: {1}".format(
                    missing_cols, extra_cols
                )
            )
            raise ClearboxWrapperException(message)
    else:
        # The model signature does not specify column names => we can only verify column count.
        if len(pdf.columns) < len(input_schema.columns):
            message = (
                "Model input is missing input columns. The model signature declares "
                "{0} input columns but the provided input only has "
                "{1} columns. Note: the columns were not named in the signature so we can "
                "only verify their count."
            ).format(len(input_schema.columns), len(pdf.columns))
            raise ClearboxWrapperException(message)
        col_names = pdf.columns[: len(input_schema.columns)]
    col_types = input_schema.column_types()
    new_pdf = pd.DataFrame()
    for i, x in enumerate(col_names):
        new_pdf[x] = _enforce_type(x, pdf[x], col_types[i])
    return new_pdf


class PyFuncModel(object):
    """
    MLflow 'python function' model.
    Wrapper around model implementation and metadata. This class is not meant to be constructed
    directly. Instead, instances of this class are constructed and returned from
    py:func:`mlflow.pyfunc.load_model`.
    ``model_impl`` can be any Python object that implements the `Pyfunc interface
    <https://mlflow.org/docs/latest/python_api/mlflow.pyfunc.html#pyfunc-inference-api>`_, and
    is returned by invoking the model's ``loader_module``.
    ``model_meta`` contains model metadata loaded from the MLmodel file.
    """

    def __init__(self, model_meta: Model, model_impl: Any):
        if not hasattr(model_impl, "predict"):
            raise ClearboxWrapperException(
                "Model implementation is missing required predict method."
            )
        if not model_meta:
            raise ClearboxWrapperException("Model is missing metadata.")
        self._model_meta = model_meta
        self._model_impl = model_impl

    def predict(self, data: PyFuncInput) -> PyFuncOutput:
        """
        Generate model predictions.
        If the model contains signature, enforce the input schema first before calling the model
        implementation with the sanitized input. If the pyfunc model does not include model
        schema, the input is passed to the model implementation as is. See `Model Signature
        Enforcement <https://www.mlflow.org/docs/latest/models.html#signature-enforcement>`_
        for more details."
        :param data: Model input
        :return: Model predictions as one of pandas.DataFrame, pandas.Series, numpy.ndarray
        or list.
        """
        input_schema = self.metadata.get_input_schema()
        if input_schema is not None:
            data = _enforce_schema(data, input_schema)
        return self._model_impl.predict(data)

    @property
    def metadata(self):
        """Model metadata."""
        if self._model_meta is None:
            raise ClearboxWrapperException("Model is missing metadata.")
        return self._model_meta

    def __repr__(self):
        info = {}
        if self._model_meta is not None:
            if (
                hasattr(self._model_meta, "run_id")
                and self._model_meta.run_id is not None
            ):
                info["run_id"] = self._model_meta.run_id
            if (
                hasattr(self._model_meta, "artifact_path")
                and self._model_meta.artifact_path is not None
            ):
                info["artifact_path"] = self._model_meta.artifact_path
            info["flavor"] = self._model_meta.flavors[FLAVOR_NAME]["loader_module"]
        return yaml.safe_dump(
            {"mlflow.pyfunc.loaded_model": info}, default_flow_style=False
        )


def load_model(model_path: str, suppress_warnings: bool = False) -> PyFuncModel:
    """Load a model that has python_function flavor.

    Parameters
    ----------
    model_path : str
        Filepath of the model directory.
    suppress_warnings : bool, optional
        If Fatal, non-fatal warning messages associated with the model loading process
        will be emitted, by default True

    Returns
    -------
    PyFuncModel
        A python_function model.

    Raises
    ------
    ClearboxWrapperException
        If the model does not have the python_function flavor.
    """
    mlmodel = Model.load(os.path.join(model_path, MLMODEL_FILE_NAME))
    pyfunc_flavor_configuration = mlmodel.flavors.get(FLAVOR_NAME)

    if pyfunc_flavor_configuration is None:
        raise ClearboxWrapperException(
            'Model does not have the "{flavor_name}" flavor'.format(
                flavor_name=FLAVOR_NAME
            )
        )

    model_python_version = pyfunc_flavor_configuration.get(PY_VERSION)

    if not suppress_warnings:
        _warn_potentially_incompatible_py_version_if_necessary(
            model_py_version=model_python_version
        )

    if CODE in pyfunc_flavor_configuration and pyfunc_flavor_configuration[CODE]:
        code_path = os.path.join(model_path, pyfunc_flavor_configuration[CODE])
        _add_code_to_system_path(code_path=code_path)

    data_path = (
        os.path.join(model_path, pyfunc_flavor_configuration[DATA])
        if (DATA in pyfunc_flavor_configuration)
        else model_path
    )

    model_implementation = importlib.import_module(
        pyfunc_flavor_configuration[MAIN]
    )._load_pyfunc(data_path)

    return PyFuncModel(model_meta=mlmodel, model_impl=model_implementation)


def _add_code_to_system_path(code_path):
    sys.path = [code_path] + _get_code_dirs(code_path) + sys.path


def _get_code_dirs(src_code_path, dst_code_path=None):
    """
    Obtains the names of the subdirectories contained under the specified source code
    path and joins them with the specified destination code path.
    :param src_code_path: The path of the source code directory for which to list
                          subdirectories.
    :param dst_code_path: The destination directory path to which subdirectory names should be
                          joined.
    """
    if not dst_code_path:
        dst_code_path = src_code_path
    return [
        (os.path.join(dst_code_path, x))
        for x in os.listdir(src_code_path)
        if os.path.isdir(x) and not x == "__pycache__"
    ]


def _warn_potentially_incompatible_py_version_if_necessary(model_py_version=None):
    """
    Compares the version of Python that was used to save a given model with the version
    of Python that is currently running. If a major or minor version difference is detected,
    logs an appropriate warning.
    """
    if model_py_version is None:
        logger.warning(
            "The specified model does not have a specified Python version. It may be"
            " incompatible with the version of Python that is currently running: Python %s",
            PYTHON_VERSION,
        )
    elif get_major_minor_py_version(model_py_version) != get_major_minor_py_version(
        PYTHON_VERSION
    ):
        logger.warning(
            "The version of Python that the model was saved in, `Python %s`, differs"
            " from the version of Python that is currently running, `Python %s`,"
            " and may be incompatible",
            model_py_version,
            PYTHON_VERSION,
        )


def save_model(
    path,
    loader_module=None,
    data_path=None,
    code_path=None,
    conda_env=None,
    mlflow_model=None,
    python_model=None,
    artifacts=None,
    signature: Signature = None,
    **kwargs
):
    """
    save_model(path, loader_module=None, data_path=None, code_path=None, conda_env=None,\
               mlflow_model=Model(), python_model=None, artifacts=None)
    Save a Pyfunc model with custom inference logic and optional data dependencies to a path on
    the local filesystem.
    For information about the workflows that this method supports, please see :ref:`"workflows
    for creating custom pyfunc models" <pyfunc-create-custom-workflows>` and
    :ref:`"which workflow is right for my use case?" <pyfunc-create-custom-selecting-workflow>`.
    Note that the parameters for the second workflow: ``loader_module``, ``data_path`` and the
    parameters for the first workflow: ``python_model``, ``artifacts``, cannot be
    specified together.
    :param path: The path to which to save the Python model.
    :param loader_module: The name of the Python module that is used to load the model
                          from ``data_path``. This module must define a method with the
                          prototype ``_load_pyfunc(data_path)``. If not ``None``,
                          this module and its dependencies must be included in one of
                          the following locations:
                          - The MLflow library.
                          - Package(s) listed in the model's Conda environment, specified by
                            the ``conda_env`` parameter.
                          - One or more of the files specified by the ``code_path`` parameter.
    :param data_path: Path to a file or directory containing model data.
    :param code_path: A list of local filesystem paths to Python file dependencies (or
                      directories containing file dependencies). These files are *prepended* to
                      the system path before the model is loaded.
    :param conda_env: Either a dictionary representation of a Conda environment or the path to a
                      Conda environment yaml file. This decsribes the environment this model
                      should be run in. If ``python_model`` is not ``None``, the Conda
                      environment must at least specify the dependencies contained in
                      :func:`get_default_conda_env()`. If ``None``, the default
                      :func:`get_default_conda_env()` environment is added to the
                      model. The following is an *example* dictionary representation of a Conda
                      environment::
                        {
                            'name': 'mlflow-env',
                            'channels': ['defaults'],
                            'dependencies': [
                                'python=3.7.0',
                                'cloudpickle==0.5.8'
                            ]
                        }
    :param mlflow_model: :py:mod:`mlflow.models.Model` configuration to which to add the
                         **python_function** flavor.
    :param python_model: An instance of a subclass of :class:`~PythonModel`. This class is
                         serialized using the CloudPickle library. Any dependencies of the class
                         should be included in one of the following locations:
                            - The MLflow library.
                            - Package(s) listed in the model's Conda environment, specified by
                              the ``conda_env`` parameter.
                            - One or more of the files specified by the ``code_path`` parameter.
                         Note: If the class is imported from another module, as opposed to being
                         defined in the ``__main__`` scope, the defining module should also be
                         included in one of the listed locations.
    :param artifacts: A dictionary containing ``<name, artifact_uri>`` entries. Remote artifact
                      URIs are resolved to absolute filesystem paths, producing a dictionary of
                      ``<name, absolute_path>`` entries. ``python_model`` can reference these
                      resolved entries as the ``artifacts`` property of the ``context``
                      parameter in :func:`PythonModel.load_context()
                      <mlflow.pyfunc.PythonModel.load_context>`
                      and :func:`PythonModel.predict() <mlflow.pyfunc.PythonModel.predict>`.
                      For example, consider the following ``artifacts`` dictionary::
                        {
                            "my_file": "s3://my-bucket/path/to/my/file"
                        }
                      In this case, the ``"my_file"`` artifact is downloaded from S3. The
                      ``python_model`` can then refer to ``"my_file"`` as an absolute filesystem
                      path via ``context.artifacts["my_file"]``.
                      If ``None``, no artifacts are added to the model.
    :param signature: (Experimental) :py:class:`Signature <mlflow.models.Signature>`
                      describes model input and output :py:class:`Schema <mlflow.types.Schema>`.
                      The model signature can be :py:func:`inferred
                      <mlflow.models.infer_signature>` from datasets with valid model input
                      (e.g. the training dataset with target column omitted) and valid model
                      output (e.g. model predictions generated on the training dataset),
                      for example:
                      .. code-block:: python
                        from mlflow.models.signature import infer_signature
                        train = df.drop_column("target_label")
                        predictions = ... # compute model predictions
                        signature = infer_signature(train, predictions)
    :param input_example: (Experimental) Input example provides one or several instances of
                          valid model input. The example can be used as a hint of what data to
                          feed the model. The given example will be converted to a Pandas
                          DataFrame and then serialized to json using the Pandas split-oriented
                          format. Bytes are base64-encoded.
    """
    mlflow_model = kwargs.pop("model", mlflow_model)
    if len(kwargs) > 0:
        raise TypeError(
            "save_model() got unexpected keyword arguments: {}".format(kwargs)
        )
    if code_path is not None:
        if not isinstance(code_path, list):
            raise TypeError(
                "Argument code_path should be a list, not {}".format(type(code_path))
            )

    first_argument_set = {
        "loader_module": loader_module,
        "data_path": data_path,
    }
    second_argument_set = {
        "artifacts": artifacts,
        "python_model": python_model,
    }
    first_argument_set_specified = any(
        [item is not None for item in first_argument_set.values()]
    )
    second_argument_set_specified = any(
        [item is not None for item in second_argument_set.values()]
    )
    if first_argument_set_specified and second_argument_set_specified:
        raise ClearboxWrapperException(
            "The following sets of parameters cannot be specified together: {first_set_keys}"
            " and {second_set_keys}. All parameters in one set must be `None`. Instead, found"
            " the following values: {first_set_entries} and {second_set_entries}".format(
                first_set_keys=first_argument_set.keys(),
                second_set_keys=second_argument_set.keys(),
                first_set_entries=first_argument_set,
                second_set_entries=second_argument_set,
            )
        )
    elif (loader_module is None) and (python_model is None):
        msg = (
            "Either `loader_module` or `python_model` must be specified. A `loader_module` "
            "should be a python module. A `python_model` should be a subclass of PythonModel"
        )
        raise ClearboxWrapperException(msg)

    if os.path.exists(path):
        raise ClearboxWrapperException("Path '{}' already exists".format(path))
    os.makedirs(path)
    if mlflow_model is None:
        mlflow_model = Model()
    if signature is not None:
        mlflow_model.signature = signature

    if first_argument_set_specified:
        return _save_model_with_loader_module_and_data_path(
            path=path,
            loader_module=loader_module,
            data_path=data_path,
            code_paths=code_path,
            conda_env=conda_env,
            mlflow_model=mlflow_model,
        )
    elif second_argument_set_specified:
        return _save_model_with_class_artifacts_params(
            path=path,
            python_model=python_model,
            artifacts=artifacts,
            conda_env=conda_env,
            code_paths=code_path,
            mlflow_model=mlflow_model,
        )


def _save_model_with_loader_module_and_data_path(
    path,
    loader_module,
    data_path=None,
    code_paths=None,
    conda_env=None,
    mlflow_model=new_model,
):
    """
    Export model as a generic Python function model.
    :param path: The path to which to save the Python model.
    :param loader_module: The name of the Python module that is used to load the model
                          from ``data_path``. This module must define a method with the
                          prototype ``_load_pyfunc(data_path)``.
    :param data_path: Path to a file or directory containing model data.
    :param code_paths: A list of local filesystem paths to Python file dependencies (or
                      directories containing file dependencies). These files are *prepended*
                      to the system path before the model is loaded.
    :param conda_env: Either a dictionary representation of a Conda environment or the path to a
                      Conda environment yaml file. If provided, this decsribes the environment
                      this model should be run in.
    :return: Model configuration containing model info.
    """

    code = None
    data = None

    if data_path is not None:
        model_file = _copy_file_or_tree(src=data_path, dst=path, dst_dir="data")
        data = model_file

    if code_paths is not None:
        for code_path in code_paths:
            _copy_file_or_tree(src=code_path, dst=path, dst_dir="code")
        code = "code"

    conda_env_subpath = "mlflow_env.yml"
    if conda_env is None:
        conda_env = get_default_conda_env()
    elif not isinstance(conda_env, dict):
        with open(conda_env, "r") as f:
            conda_env = yaml.safe_load(f)
    with open(os.path.join(path, conda_env_subpath), "w") as f:
        yaml.safe_dump(conda_env, stream=f, default_flow_style=False)

    add_pyfunc_flavor_to_model(
        mlflow_model,
        loader_module=loader_module,
        code=code,
        data=data,
        env=conda_env_subpath,
    )
    mlflow_model.save(os.path.join(path, MLMODEL_FILE_NAME))
    return mlflow_model



from sys import version_info
from typing import Dict, List, Optional, Union

import yaml

CONDA_HEADER = """\
name: conda-env
channels:
  - defaults
  - conda-forge
"""

PYTHON_VERSION = "{major}.{minor}.{micro}".format(
    major=version_info.major, minor=version_info.minor, micro=version_info.micro
)


def get_major_minor_py_version(py_version):
    return ".".join(py_version.split(".")[:2])


def _get_default_conda_env(
    path: str = None,
    additional_conda_deps: Optional[List] = None,
    additional_pip_deps: Optional[List] = None,
    additional_conda_channels: Optional[List] = None,
) -> Union[Dict, None]:
    """Generate and, optionally, save to file the default Conda environment for models.

    Parameters
    ----------
    path : str, optional
        File path. If not None, the Conda env will be saved to file, by default None
    additional_conda_deps : Optional[List], optional
        List of additional conda dependencies, by default None
    additional_pip_deps : Optional[List], optional
        List of additional Pypi dependencies, by default None
    additional_conda_channels : Optional[List], optional
        List of additional conda channels, by default None

    Returns
    -------
    Union[Dict, None]
        None if path is not None, else the Conda environment generated as a dictionary.
    """
    pip_deps = additional_pip_deps if additional_pip_deps else []
    conda_deps = (additional_conda_deps if additional_conda_deps else []) + (
        ["pip"] if pip_deps else []
    )

    env = yaml.safe_load(CONDA_HEADER)
    env["dependencies"] = ["python={}".format(PYTHON_VERSION)]
    if conda_deps is not None:
        env["dependencies"] += conda_deps
    env["dependencies"].append({"pip": pip_deps})
    if additional_conda_channels is not None:
        env["channels"] += additional_conda_channels

    if path is not None:
        with open(path, "w") as out:
            yaml.safe_dump(env, stream=out, default_flow_style=False)
        return None
    else:
        return env



from .utils import add_clearbox_flavor_to_model, FLAVOR_NAME
from .wrapper import load_model, save_model


__all__ = [
    add_clearbox_flavor_to_model,
    FLAVOR_NAME,
    load_model,
    save_model,
]



from datetime import datetime
import json
import os
from typing import Dict, Optional

import yaml

from clearbox_wrapper.schema import Schema
from clearbox_wrapper.signature import Signature


MLMODEL_FILE_NAME = "MLmodel"


class Model(object):
    """A ML Model that can support multiple model flavors."""

    def __init__(
        self,
        timestamp: Optional[datetime] = None,
        flavors: Optional[Dict] = None,
        model_signature: Optional[Signature] = None,
        preprocessing_signature: Optional[Signature] = None,
        preparation_signature: Optional[Signature] = None,
    ) -> None:
        """Create a new Model object.

        Parameters
        ----------
        timestamp : Optional[datetime], optional
            A timestamp of the model creation, by default None.
            If None, it will be use datetime.utcnow()
        flavors : Optional[Dict], optional
            Dictionary of flavors: a "flavor" is a convention
            representing the framework the model was created with
            and/or downstream tools that can "understand" the model,
            by default None.
        signature : Optional[Signature], optional
            Description of the model inputs as Signature oject,
            by default None
        """
        self.timestamp = str(timestamp or datetime.utcnow())
        self.flavors = flavors if flavors is not None else {}
        self.model_signature = model_signature
        self.preprocessing_signature = preprocessing_signature
        self.preparation_signature = preparation_signature

    def __eq__(self, other: "Model") -> bool:
        """Check if two models are equal.

        Parameters
        ----------
        other : Model
            A Model object

        Returns
        -------
        bool
            true, if the other Model is equal to self.
        """
        if not isinstance(other, Model):
            return False
        return self.__dict__ == other.__dict__

    def get_data_preparation_input_schema(self) -> Optional[Schema]:
        """Get the data preparation inputs schema."

        Returns
        -------
        Schema
            Data preparation inputs schema: specification of types and column names.
        """
        return (
            self.preparation_signature.inputs
            if self.preparation_signature is not None
            else None
        )

    def get_data_preparation_output_schema(self) -> Optional[Schema]:
        """Get the data preparation outputs schema."

        Returns
        -------
        Schema
            Data preparation outputs schema: specification of types and column names.
        """
        return (
            self.preparation_signature.outputs
            if self.preparation_signature is not None
            else None
        )

    def get_preprocessing_input_schema(self) -> Optional[Schema]:
        """Get the preprocessing inputs schema."

        Returns
        -------
        Schema
            Preprocessing inputs schema: specification of types and column names.
        """
        return (
            self.preprocessing_signature.inputs
            if self.preprocessing_signature is not None
            else None
        )

    def get_preprocessing_output_schema(self) -> Optional[Schema]:
        """Get the preprocessing outputs schema."

        Returns
        -------
        Schema
            Preprocessing outputs schema: specification of types and column names.
        """
        return (
            self.preprocessing_signature.outputs
            if self.preprocessing_signature is not None
            else None
        )

    def get_model_input_schema(self) -> Optional[Schema]:
        """Get the model inputs schema."

        Returns
        -------
        Schema
            Model inputs schema: specification of types and column names.
        """
        return self.model_signature.inputs if self.model_signature is not None else None

    def get_model_output_schema(self) -> Optional[Schema]:
        """Get the model outputs schema."

        Returns
        -------
        Schema
            Model outputs schema: specification of types and column names.
        """
        return (
            self.model_signature.outputs if self.model_signature is not None else None
        )

    def add_flavor(self, name: str, **params) -> "Model":
        """Add an entry for how to serve the model in a given format.

        Returns
        -------
        Model
            self
        """
        self.flavors[name] = params
        return self

    @property
    def preparation_signature(self) -> Optional[Signature]:
        """Get the data preparation signature associated with the model.

        Returns
        -------
        Optional[Signature]
            Data preparation signature as a Signature instance.
        """
        return self._preparation_signature

    @preparation_signature.setter
    def preparation_signature(self, value: Signature) -> None:
        """Set the data preparation signature of the model

        Parameters
        ----------
        value : Signature
            Data preparation signature as a Signature instance.
        """
        self._preparation_signature = value

    @property
    def preprocessing_signature(self) -> Optional[Signature]:
        """Get the preprocessing signature associated with the model.

        Returns
        -------
        Optional[Signature]
            Preprocessing signature as a Signature instance.
        """
        return self._preprocessing_signature

    @preprocessing_signature.setter
    def preprocessing_signature(self, value: Signature) -> None:
        """Set the preprocessing signature of the model

        Parameters
        ----------
        value : Signature
            Preprocessing signature as a Signature instance.
        """
        self._preprocessing_signature = value

    @property
    def model_signature(self) -> Optional[Signature]:
        """Get the model signature.

        Returns
        -------
        Optional[Signature]
            The model signature: it defines the schema of a model inputs and outputs.
            Model inputs/outputs are described as a sequence of (optionally) named
            columns with type.
        """
        return self._model_signature

    @model_signature.setter
    def model_signature(self, value: Signature) -> None:
        """Set the model signature

        Parameters
        ----------
        value : Signature
            Model signature as a Signature object.
        """
        self._model_signature = value

    def to_dict(self) -> Dict:
        """Get the model attributes as a dictionary.

        Returns
        -------
        Dict
            Model attributes as dict <attribute_name: attribute_value>
        """
        model_dict = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        if self.model_signature is not None:
            model_dict["signature"] = self.model_signature.to_dict()
        if self.preprocessing_signature is not None:
            model_dict[
                "preprocessing_signature"
            ] = self.preprocessing_signature.to_dict()
        if self.preparation_signature is not None:
            model_dict["preparation_signature"] = self.preparation_signature.to_dict()
        return model_dict

    def to_yaml(self, stream=None):
        """Serialize model object into a YAML stream. If stream is None,
        return the produced string instead.

        Parameters
        ----------
        stream : f, optional
            YAML stream, by default None

        Returns
        -------
        str
            YAML stream. If stream is None, return the produced string instead.
        """
        return yaml.safe_dump(self.to_dict(), stream=stream, default_flow_style=False)

    def __str__(self) -> str:
        """Get model representation as string.

        Returns
        -------
        str
            Model attributes as string.
        """
        return self.to_yaml()

    def to_json(self) -> str:
        """Get model representation in JSON format.

        Returns
        -------
        str
            Model attributes in JSON format.
        """
        return json.dumps(self.to_dict())

    def save(self, path: str) -> None:
        """Save model YAML representation to a file

        Parameters
        ----------
        path : str
            Path of the file to save the YAML in.
        """
        with open(path, "w") as out:
            self.to_yaml(out)

    @classmethod
    def load(cls, path: str) -> "Model":
        """Load a model from file.

        Parameters
        ----------
        path : str
            Path of the file to load the Model from.

        Returns
        -------
        Model
            Loaded Model.
        """
        if os.path.isdir(path):
            path = os.path.join(path, MLMODEL_FILE_NAME)
        with open(path) as f:
            return cls.from_dict(yaml.safe_load(f.read()))

    @classmethod
    def from_dict(cls, model_dict: Dict) -> "Model":
        """Load a model from its YAML representation.

        Parameters
        ----------
        model_dict : Dict
            Model dictionary representation.

        Returns
        -------
        Model
            Loaded Model.
        """

        if "signature" in model_dict and isinstance(model_dict["signature"], dict):
            model_dict = model_dict.copy()
            model_dict["model_signature"] = Signature.from_dict(model_dict["signature"])
            del model_dict["signature"]

        if "preprocessing_signature" in model_dict and isinstance(
            model_dict["preprocessing_signature"], dict
        ):
            model_dict = model_dict.copy()
            model_dict["preprocessing_signature"] = Signature.from_dict(
                model_dict["preprocessing_signature"]
            )

        if "preparation_signature" in model_dict and isinstance(
            model_dict["preparation_signature"], dict
        ):
            model_dict = model_dict.copy()
            model_dict["preparation_signature"] = Signature.from_dict(
                model_dict["preparation_signature"]
            )

        return cls(**model_dict)



from typing import Any, Dict, Union

import numpy as np
import pandas as pd

from clearbox_wrapper.schema import _infer_schema, Schema

InferableDataset = Union[pd.DataFrame, np.ndarray, Dict[str, np.ndarray]]


class Signature(object):
    def __init__(self, inputs: Schema, outputs: Schema = None):
        if not isinstance(inputs, Schema):
            raise TypeError("inputs must be type Schema, got '{}'".format(type(inputs)))
        if outputs is not None and not isinstance(outputs, Schema):
            raise TypeError(
                "outputs must be either None or mlflow.models.signature.Schema, "
                "got '{}'".format(type(inputs))
            )
        self.inputs = inputs
        self.outputs = outputs

    def to_dict(self) -> Dict[str, Any]:
        return {
            "inputs": self.inputs.to_json(),
            "outputs": self.outputs.to_json() if self.outputs is not None else None,
        }

    @classmethod
    def from_dict(cls, signature_dict: Dict[str, Any]):
        """
        Deserialize from dictionary representation.
        :param signature_dict: Dictionary representation of model signature.
                               Expected dictionary format:
                               `{'inputs': <json string>}`
        :return: Signature populated with the data form the dictionary.
        """
        inputs = Schema.from_json(signature_dict["inputs"])
        if "outputs" in signature_dict and signature_dict["outputs"] is not None:
            outputs = Schema.from_json(signature_dict["outputs"])
            return cls(inputs, outputs)
        else:
            return cls(inputs)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Signature)
            and self.inputs == other.inputs
            and self.outputs == other.outputs
        )

    def __repr__(self) -> str:
        return (
            "inputs: \n"
            "  {}\n"
            "outputs: \n"
            "  {}\n".format(repr(self.inputs), repr(self.outputs))
        )


def infer_signature(input_data: Any, output_data: InferableDataset = None) -> Signature:
    """
    Infer an MLflow model signature from the training data (input).
    The signature represents model input as data frames with (optionally) named columns
    and data type specified as one of types defined in :py:class:`mlflow.types.DataType`.
    This method will raise an exception if the user data contains incompatible types or is not
    passed in one of the supported formats listed below.
    The input should be one of these:
      - pandas.DataFrame
      - dictionary of { name -> numpy.ndarray}
      - numpy.ndarray
      - pyspark.sql.DataFrame
    The element types should be mappable to one of :py:class:`mlflow.types.DataType`.
    NOTE: Multidimensional (>2d) arrays (aka tensors) are not supported at this time.
    :param model_input: Valid input to the model. E.g. (a subset of) the training dataset.
    :param model_output: Valid model output. E.g. Model predictions for the (subset of) training
                         dataset.
    :return: Signature
    """
    inputs = _infer_schema(input_data)
    outputs = _infer_schema(output_data) if output_data is not None else None
    return Signature(inputs, outputs)
